module Parallel where

import Control.Monad (guard)
import Control.Monad.Loops
import Data.Sequence (Seq, (><), (|>), (<|))
import qualified Data.Sequence as Seq
import Database.HDBC
import Database.HDBC.Sqlite3
import DB
import Utils (seqLast)
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Control.Concurrent
import Control.Concurrent.Chan
import Control.Parallel
import Data.Foldable (toList)

data ParState = ParState {
  seen :: MVar (Set Int),
  target :: Int,
  conn :: IO Connection,
  queue :: Chan (Seq Int),
  resultChan :: Chan (Seq Int)
                   }

parbfs :: (Int, Int) -> Int -> IO (Seq Int)
parbfs (source, target) numWorkers = do
  resultChan <- newChan
  queueChan <- newChan
  seenVar <- newEmptyMVar
  putMVar seenVar Set.empty
  let conn = connectSqlite3 "data/sixdegrees.sqlite"
      initialQueue = Seq.singleton source
      initialState = ParState {queue=queueChan, seen=seenVar, resultChan=resultChan, target=target, conn=conn}
  writeChan queueChan initialQueue
  mapM_ (\state -> do
            withConn <- createConn state
            forkIO (handleBfs withConn)) $ take numWorkers $ repeat initialState
  final <- readChan resultChan
  return $ final
  where createConn (ParState {resultChan=result,queue=queue, target=target,seen=seen, conn=conn}) = do
          unwrapped <- conn
          return ParState {resultChan=result,queue=queue, target=target,seen=seen, conn=(clone unwrapped)}


handleBfs :: ParState -> IO ()
handleBfs state = do
  doBfs state
  where
    doBfs (ParState {resultChan=resultChan, queue=queue, conn=conn, seen=seen, target=t}) = do
          path <- readChan queue
          guard (Seq.length path < 7) >> return ()
          lastSeen <- takeMVar seen
          let curr = seqLast path
              newSeen = Set.insert curr lastSeen
          putMVar seen newSeen
          nextSteps <- getOutgoingLinksForPage conn curr
          writeList2Chan queue $ toList (getNextQueuePaths path nextSteps newSeen)
          if curr == t || (elem t nextSteps)
            then writeChan resultChan (path |> t)
            else doBfs state
    getNextQueuePaths base steps seen = foldl (\acc x -> if Set.member x seen then acc else acc |> (base |> x)) Seq.empty steps
