module Serial where

import Control.Monad.State
import Control.Monad.Loops
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set
import Data.Sequence (Seq, (><), (|>), (<|))
import qualified Data.Sequence as Seq
import Database.HDBC
import Database.HDBC.Sqlite3
import DB
import Utils (seqLast)

data BFSState = BFSState {
  queue :: Seq (Seq Int),
  seen :: Set Int,
  target :: Int,
  conn :: IO Connection
                   }

bfs :: Int -> Int -> IO (Seq Int)
bfs source target = do
  let conn = connectSqlite3 "data/sixdegrees.sqlite"
      initialQueue = Seq.singleton $ Seq.singleton source
      initialState = BFSState {queue=initialQueue, conn=conn, target=target, seen=Set.empty}
  (BFSState {queue=final}) <- iterateUntilM atDepthLimit doBfs initialState
  return $ Seq.index final 0
  where atDepthLimit (BFSState {queue=queue, conn=_, target=_, seen=_}) = do
          let first = Seq.index queue 0
              depth = Seq.length first
            in depth > 7 || (queue == Seq.empty) || (isResult queue)
        isResult queue = Seq.length queue == 1 && seqLast (Seq.index queue 0) == target

doBfs :: BFSState -> IO BFSState
doBfs (BFSState {queue=queue, conn=conn, target=target, seen=seen})= do
  let path = Seq.index queue 0
      curr = seqLast path
      newQueue = Seq.drop 1 queue
      newSeen = Set.insert curr seen
  nextSteps <- getOutgoingLinksForPage conn curr
  if curr == target || (elem target nextSteps)
    then return $ BFSState {queue=Seq.singleton (path |> target), conn=conn, target=target, seen=newSeen}
    else return $ BFSState {queue=queue >< (getNextQueuePaths path nextSteps newSeen), conn=conn, target=target, seen=newSeen}
  where getNextQueuePaths base steps seen = foldl (\acc x -> if Set.member x seen then acc else acc |> (base |> x)) Seq.empty steps

{-
bfs :: Int -> Int -> IO (Seq Int)
bfs source target = do
  let conn = connectSqlite3 "data/sdow.sqlite"
  put (conn, target, Set.empty)
  result <- doBfs $ Seq.singleton $ Seq.singleton source
  let final = until atDepthLimit doBfs $ Seq.singleton $ Seq.singleton source
  return $ Seq.index result 0
  where atDepthLimit wrapped = do
          let first = Seq.index queue 0
              depth = Seq.length first
            in depth > 7 || (queue == Seq.empty) || (isResult queue)
        isResult queue = Seq.length queue == 1 && seqLast (Seq.index queue 0) == target


-- Run until state has result
-- current state has form (path queue, seen, depth)
-- When done, puts result
doBfs :: IO (Seq (Seq Int)) -> IO (Seq (Seq Int))
doBfs wrapped = do
  (conn, target, seen) <- get
  queue <- wrapped
  let path = Seq.index queue 0
      curr = seqLast path
      newQueue = Seq.drop 1 queue
      newSeen = Set.insert curr seen
  guard (curr == target) >> return $ Seq.singleton (path |> target)
  nextSteps <- getOutgoingLinksForPage conn curr
  guard (elem target nextSteps) >> return $ Seq.singleton (path |> target)
  put (conn, target, newSeen)
  return $ doBfs $ queue >< (getNextQueuePaths path nextSteps newSeen)
  where getNextQueuePaths base steps seen = foldl (\acc x -> if Set.member x seen then acc else acc |> (base |> x)) Seq.empty steps
-}

{-
bfs :: (Int, Int) -> IO (Seq Int)
bfs (source, target) =
  let conn = connectSqlite3 "data/sdow.sqlite"
      queue = Seq.singleton $ Seq.singleton source
      seen = Set.empty
  in doBfs conn queue seen
  where doBfs conn queue seen = do
          let path = Seq.index queue 0
              curr = seqLast path
              newQueue = Seq.drop 1 queue
              newSeen = Set.insert curr seen
          guard (curr==target) >> return $ Seq.singleton (path |> target)
          nextSteps <- getOutgoingLinksForPage conn curr
          guard (elem target nextSteps) >> return $ Seq.singleton (path |> target)
          return $ doBfs conn (queue >< (getNextQueuePaths path nextSteps newSeen)) newSeen
        getNextQueuePaths base steps seen = foldl (\acc x -> if Set.member x seen then acc else acc |> (base |> x)) Seq.empty steps
-}



{-
bfs :: (Int, Int) -> IO (Seq Int)
bfs (source, target) =
  let conn = connectSqlite3 "data/sdow.sqlite"
      state = BFSState { queue=(Seq.singleton $ Seq.singleton source), seen=Set.empty, target=target, conn=conn}

doBfs :: BFSState -> IO BFSState
doBfs (BFSState {queue = queue, seen = seen, target=target, conn=conn}) = do
  let path = Seq.index queue 0
      curr = seqLast path
      newQueue = Seq.drop 1 queue
      newSeen = Set.insert curr seen
  guard (curr == target) >> return $ BFSState {queue=(Seq.singleton $ path |> target), seen=newSeen, conn=conn, target=target}
  nextSteps <- getOutgoingLinksForPage conn curr
  guard (elem target nextSteps) >> return $ BFSState {queue=(Seq.singleton $ path |> target), seen=newSeen, conn=conn, target=target}
  let toAdd = getNextQueuePaths path nextSteps newSeen
  return $ BFSState {queue=(queue >< toAdd), seen=newSeen, conn=conn, target=target}
  where getNextQueuePaths base steps seen = foldl (\acc x -> if Set.member x seen then acc else acc |> (base |> x)) Seq.empty steps
-}


 {-
data BiState = BiState {
  unvisitedForward :: Map Int (Seq Int),
  unvisitedBackward :: Map Int (Seq Int),
  visitedForward :: Map Int (Seq Int),
  vistedBackward :: Map Int (Seq Int),
  forwardDepth :: Int,
  backwardDepth :: Int,
  paths :: Seq (Seq Int),
  conn = IO Connection
                       }

initialState :: Int -> Int -> IO Connection -> BiState
initialState sourceID targetID conn= BiState {
  source = sourceID,
  target = targetID,
  unvisitedForward = Map.fromList [(sourceID, 0)],
  unvisitedBackward = Map.fromList [(targetID, 0)],
  visitedForward = Map.empty,
  visitedBackward = Map.empty,
  forwardDepth = 0,
  backwardDepth = 0,
  paths = Seq.empty,
  conn=conn
                       }

biBfs :: Int -> Int -> IO (Seq Int)
biBfs source target = do
  let conn = connectSqlite3 "data/sdow.sqlite"
      init = initialState source target conn
  resultState <- iterateUntilM isFinished doBiBfs init
where isFinished (BiState {paths=paths, unvisitedForward = uf, unvisitedBackward=ub}) = Seq.length paths > 0 || Map.size uf == 0 || Map.size ub == 0

doBiBfs BiState -> IO BiState
doBiBfs state = do
  fCount <- getOutgoingLinksCounts conn (Map.keys uf)
  bCount <- getIncomingLinksCounts conn (Map.keys ub)

  if fCount < bCount then runForward state else runBackward state
    where runForward (BiState {conn=conn, paths=ps, visitedForward=vf, visitedBackward=vb, unvisitedForward=uf, unvisitedBackward=ub, forwardDepth = fDepth, backwardDepth=bDepth, source=source, target=target }) = do
          let newFDepth = fDepth + 1
          outgoingLinks <- getOutgoingLinks conn (Map.keys uf)
          let newUf = Map.empty
              newVf = foldl (\acc key -> case Map.lookup key uf of
                            Just x -> Map.insert key x acc
                            Nothing -> acc) vf $ Map.keys uf
          map (updateUnvisited uf sourcePage) outgoingLinks
          foldl (\acc x -> updateUnvisited acc v source x)
        where updateUnvisited u v source target = if Map.notMember target v && Map.notMember target u
                then Map.insert target $ Seq.singleton source
                else if Map.member target u then Map.update (\x -> x |> source) target u else u

updateUnvisited :: Map Int (Seq Int) -> Int -> Int -> Map Int (Seq Int)
updateUnvisited u source target = if


getPaths :: Seq Int -> Map Int Int -> Seq (Seq Int)
getPaths pageIDs visited = foldl (\acc x -> if x == 0
                                   then (acc |> Seq.empty)
                                   else case Map.lookup visited x of
                                     Just val -> acc >< $ buildCurrentPaths x (getPaths (Seq.singleton val) visited)
                                     Nothing -> acc) Seq.empty pageIDs
  where buildCurrentPaths pageID paths = foldl (\acc path -> acc |> (path |> pageID)) Seq.empty paths

-}
